#include  <math.h>
#include "millis.h"
#include "servo.h"

#define DEG2RAD(g) ((g)*3.141592)/180.

//-- Oscillators parameters    
uint16_t A[4];          //-- Amplitude (degrees)
uint16_t O[4];          //-- Offset (degrees)
double phase_diff[4];   //-- Phase (radians): correspond to _phase0

//-- Internal variables
int16_t pos[4];                     //-- Current servo pos
int16_t trim[4] = {7, 8, -5, -6};   //-- Calibration offset (degree)
double phase[4];                    //-- Current phase     (radian)
double inc;                         //-- Increment of phase (radian)
uint16_t TS = 20;                   //-- sampling period (msec)

uint32_t currentMillis[4];
uint32_t previousMillis[4];

bool stop = false;
bool rev[4];

double round_d(double x) {
    if (x > 0.0) {
        return floor(x + 0.5);
    } else {
        return -1.0 * floor(fabs(x) + 0.5);
    }
}

//-- This function returns true if another sample
//-- should be taken (i.e. the TS time has passed since
//-- the last sample was taken

bool next_sample(uint8_t servo) {

    //-- Read current time
    currentMillis[servo] = millis();

    //-- Check if the timeout has passed
    if (currentMillis[servo] - previousMillis[servo] >= TS) {
        //  if(_currentMillis - _previousMillis > _TS) {// this is a bug!
        previousMillis[servo] = currentMillis[servo];

        return true;
    }

    return false;
}

/*******************************************************************/
/* This function should be periodically called                     */
/* in order to maintain the oscillations. It calculates            */
/* if another sample should be taken and position the servo if so  */

/*******************************************************************/
void refresh(int servo) {

    //-- Only When TS milliseconds have passed, the new sample is obtained
    if (next_sample(servo)) {

        //-- If the oscillator is not stopped, calculate the servo position
        if (!stop) {
            //-- Sample the sine function and set the servo pos
            pos[servo] = (int16_t) round_d(A[servo] * sin(phase[servo] + phase_diff[servo]) + O[servo]);
            if (rev[servo]) pos[servo] = -pos[servo];
            //  _servo.write(_pos+90+_trim);

            if (servo == 0) OC1R = ANG2VAL(pos[servo] + 90 + trim[servo]);
            else if (servo == 1) OC2R = ANG2VAL(pos[servo] + 90 + trim[servo]);
            else if (servo == 2) OC3R = ANG2VAL(pos[servo] + 90 + trim[servo]);
            else if (servo == 3) OC4R = ANG2VAL(pos[servo] + 90 + trim[servo]);
        }

        //-- Increment the phase
        //-- It is always increased, even when the oscillator is stop
        //-- so that the coordination is always kept
        phase[servo] = phase[servo] + inc;

    }
}

void oscillateServos(uint16_t T, double cycle) {

    inc = 2 * 3.14159265 / T*TS;

    double ref = (double) millis();
    double x;
    int i;

    for (x = ref; x <= (double) T * cycle + ref; x = (double) millis()) {
        for (i = 0; i < 4; i++) {
            refresh(i);
        }
    }
}

void _execute(uint16_t T, double steps) {

    uint16_t cycles = (uint16_t) steps;
    int i;

    //-- Execute complete cycles
    if (cycles >= 1)
        for (i = 0; i < cycles; i++)
            oscillateServos(T, 1.0);

    //-- Execute the final not complete cycle    
    oscillateServos(T, (double) (steps - cycles));
}

//---------------------------------------------------------
//-- Otto gait: Walking  (forward or backward)    
//--  Parameters:
//--    * steps:  Number of steps
//--    * T : Period (miliseconds)
//--    * Dir: Direction: FORWARD / BACKWARD
//---------------------------------------------------------

void walk(double steps, uint16_t T, uint16_t dir) {

    //-- Oscillator parameters for walking
    //-- Hip sevos are in phase
    //-- Feet servos are in phase
    //-- Hip and feet are 90 degrees out of phase
    //--      -90 : Walk forward
    //--       90 : Walk backward
    //-- Feet servos also have the same offset (for tiptoe a little bit)

    A[0] = 30;
    A[1] = 30;
    A[2] = 20;
    A[3] = 20;
    O[0] = 0;
    O[1] = 0;
    O[2] = 4;
    O[3] = -4;
    phase_diff[0] = 0;
    phase_diff[1] = 0;
    phase_diff[2] = DEG2RAD(dir * -90);
    phase_diff[3] = DEG2RAD(dir * -90);

    //-- Let's oscillate the servos!
    _execute(T, steps);

}


//---------------------------------------------------------
//-- Otto gait: Moonwalker. Otto moves like Michael Jackson
//--  Parameters:
//--    Steps: Number of steps
//--    T: Period (miliseconds)
//--    h: Height. Typical valures between 15 and 40
//--    dir: Direction: LEFT / RIGHT
//---------------------------------------------------------

void moonwalker(double steps, uint16_t T, uint16_t h, uint16_t dir) {

    //-- This motion is similar to that of the caterpillar robots: A travelling
    //-- wave moving from one side to another
    //-- The two Otto's feet are equivalent to a minimal configuration. It is known
    //-- that 2 servos can move like a worm if they are 120 degrees out of phase
    //-- In the example of Otto, the two feet are mirrored so that we have:
    //--    180 - 120 = 60 degrees. The actual phase difference given to the oscillators
    //--  is 60 degrees.
    //--  Both amplitudes are equal. The offset is half the amplitud plus a little bit of
    //-   offset so that the robot tiptoe lightly

    uint16_t phi = -dir * 90;

    A[0] = 0;
    A[1] = 0;
    A[2] = h;
    A[3] = h;
    O[0] = 0;
    O[1] = 0;
    O[2] = h / 2 + 2;
    O[3] = -h / 2 - 2;
    phase_diff[0] = 0;
    phase_diff[1] = 0;
    phase_diff[2] = DEG2RAD(phi);
    phase_diff[3] = DEG2RAD(-60 * dir + phi);

    //-- Let's oscillate the servos!
    _execute(T, steps);
}
